/* 
   Unix SMB/CIFS implementation.

   AIX loadable authentication module, providing identification 
   routines against Samba winbind/Windows NT Domain

   Copyright (C) Aaron Collins 2003
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.   
*/

#include "winbind_client.h"
#include <nsswitch.h>
#include <iaf.h>
#include <ia.h>


#define NSDB_GROUP		"group"
#define NSDB_PASSWD		"passwd"

#define NSS_METHOD_PROTOTYPE(method) \
	int method(void *, void *, va_list)

/* Make sure that the module gets registered needed by freebsd 5.1 */

extern enum nss_status _nss_winbind_getgrent_r(struct group *, char *, size_t,
    int *);
extern enum nss_status _nss_winbind_getgrnam_r(const char *, struct group *,
    char *, size_t, int *);
extern enum nss_status _nss_winbind_getgrgid_r(gid_t gid, struct group *, char *,
    size_t, int *);
extern enum nss_status _nss_winbind_setgrent(void);
extern enum nss_status _nss_winbind_endgrent(void);

extern enum nss_status _nss_winbind_getpwent_r(struct passwd *, char *, size_t,
    int *);
extern enum nss_status _nss_winbind_getpwnam_r(const char *, struct passwd *,
    char *, size_t, int *);
extern enum nss_status _nss_winbind_getpwuid_r(gid_t gid, struct passwd *, char *,
    size_t, int *);
extern enum nss_status _nss_winbind_setpwent(void);
extern enum nss_status _nss_winbind_endpwent(void);

NSS_METHOD_PROTOTYPE(__nss_compat_getgrnam_r);
NSS_METHOD_PROTOTYPE(__nss_compat_getgrgid_r);
NSS_METHOD_PROTOTYPE(__nss_compat_getgrent_r);
NSS_METHOD_PROTOTYPE(__nss_compat_setgrent);
NSS_METHOD_PROTOTYPE(__nss_compat_endgrent);

NSS_METHOD_PROTOTYPE(__nss_compat_getpwnam_r);
NSS_METHOD_PROTOTYPE(__nss_compat_getpwuid_r);
NSS_METHOD_PROTOTYPE(__nss_compat_getpwent_r);
NSS_METHOD_PROTOTYPE(__nss_compat_setpwent);
NSS_METHOD_PROTOTYPE(__nss_compat_endpwent);

NSS_METHOD_PROTOTYPE(_nss_winbind_openinfo);

static ns_mtab methods[] = {
{ NSDB_GROUP, "getgrnam_r", __nss_compat_getgrnam_r, (void *) _nss_winbind_getgrnam_r },
{ NSDB_GROUP, "getgrgid_r", __nss_compat_getgrgid_r, (void *) _nss_winbind_getgrgid_r },
{ NSDB_GROUP, "getgrent_r", __nss_compat_getgrent_r, (void *) _nss_winbind_getgrent_r },
{ NSDB_GROUP, "setgrent",   __nss_compat_setgrent,   (void *) _nss_winbind_setgrent },
{ NSDB_GROUP, "endgrent",   __nss_compat_endgrent,   (void *) _nss_winbind_endgrent },

{ NSDB_PASSWD, "getpwnam_r", __nss_compat_getpwnam_r, (void *) _nss_winbind_getpwnam_r },
{ NSDB_PASSWD, "getpwuid_r", __nss_compat_getpwuid_r, (void *) _nss_winbind_getpwuid_r },
{ NSDB_PASSWD, "getpwent_r", __nss_compat_getpwent_r, (void *) _nss_winbind_getpwent_r },
{ NSDB_PASSWD, "setpwent",   __nss_compat_setpwent,   (void *) _nss_winbind_setpwent },
{ NSDB_PASSWD, "endpwent",   __nss_compat_endpwent,   (void *) _nss_winbind_endpwent },

{ "iaf",	"ia_openinfo",	_nss_winbind_openinfo, 0},
};

ns_mtab *
nss_module_register(const char *source, unsigned int *mtabsize,
    nss_module_unregister_fn *unreg)
{
        *mtabsize = sizeof(methods)/sizeof(methods[0]);
        *unreg = NULL;
        return (methods);
}

/* int ia_openinfo(const char  * logname, uinfo_t  * uinfo); */
int 
_nss_winbind_openinfo(void *rv, void *mdata, va_list ap)
{
	const char *name;
	struct passwd p;
	struct group grp, *gr;
	char buf[512], grbuf[128];
	enum nss_status status;
	int err, dlen, slen, grlen;
	gid_t *gp, smlist[32]; /* usually long enough */
	int *ret = rv;
	uinfo_t *uipp;
	char *grptr;
	uinfo_t uip;
	long n, max;

	name = va_arg(ap, const char *);
	uipp = va_arg(ap, uinfo_t *);

	*uipp = 0;
	
	status = _nss_winbind_getpwnam_r(name, &p, buf, 512, &err);
	status = __nss_compat_result(status, err);

	if (status != NS_SUCCESS) {
		return status;
	}

	/* Fill gp[] with all GIDs associated with p.pw_name. */
	if ((max = sysconf(_SC_NGROUPS_MAX)) <= 0) {
	usesm:;
		max = sizeof(smlist) / sizeof(gid_t);
		gp = smlist;
	} else if (max <= sizeof(smlist) / sizeof(gid_t)) {
		gp = smlist;
	} else if ((gp = malloc(max * sizeof(gid_t))) == 0) {
		goto usesm;
	}
	n = 0;
	if (p.pw_gid >= 0)
		gp[n++] = p.pw_gid;
	setgrent();
	grptr = grbuf;
	grlen = sizeof(grbuf);
	/*
	 * Walk through the /etc/group file (and other module's
	 * variants) filling in the supplemental GIDs.  Loop here
	 * until reach EOF or, when the buffer is too short,
	 * retry the current entry.
	 */
	for (;;) {
		err = getgrent_r(&grp, grptr, grlen, &gr);
		if (gr != 0) {
			if (n >= max) /* no more room for GIDs */
				break;
			if (grp.gr_gid == p.pw_gid) /* skip primary group */
				continue;
			while (*grp.gr_mem != 0) {
				if (strcmp(*grp.gr_mem, p.pw_name) == 0) {
					gp[n++] = grp.gr_gid;
					break;
				}
				grp.gr_mem++;
			}
			continue;
		}
		if (err == ERANGE) {
			if (grptr != grbuf)
				free(grptr);
			grlen <<= 1;
			if ((grptr = malloc(grlen)) != 0)
				continue;
		}
		break;
	}
	endgrent();
	if (err != 0) {
		status = NS_TRYAGAIN;
		goto out;
	}

	/* Now calculate the space needed for result struct */
	max = n;
	dlen = strlen(p.pw_dir);
	slen = strlen(p.pw_shell);

	/* struct + max groups + dir + shell */
	n = sizeof(*uip) + max * sizeof(gid_t) + dlen + 1 + slen + 1;

	if ((uip = malloc(n)) == NULL) {
		status = NS_TRYAGAIN;
		goto out;
	}

	bzero(uip, n);

	strncpy(uip->ia_name, p.pw_name, sizeof(uip->ia_name));
	strncpy(uip->ia_pwdp, p.pw_passwd, sizeof(uip->ia_pwdp));
	uip->ia_uid = p.pw_uid;
	uip->ia_gid = p.pw_gid;
	uip->ia_lstchg = -1;
	uip->ia_min = -1;
	uip->ia_max = -1;
	uip->ia_warn = -1;
	uip->ia_inact = -1;
	uip->ia_expire = -1;
	uip->ia_flag = 0;
	uip->ia_dirsz = dlen;
	uip->ia_shsz = slen;
	uip->ia_lvlcnt = 0;
	uip->ia_sgidcnt = max;

	uip->ia_lvlp = (level_t *)&uip[1];
	uip->ia_sgidp = (gid_t *)&uip[1];
	uip->ia_dirp = (char *)&uip->ia_sgidp[max];
	uip->ia_shellp = &uip->ia_dirp[dlen + 1];
	memcpy(uip->ia_sgidp, gp, max * sizeof(gid_t));
	strcpy(uip->ia_dirp, p.pw_dir);
	strcpy(uip->ia_shellp, p.pw_shell);

	*ret = 0;
	*uipp = uip;
out:;
	if (gp != smlist)
		free(gp);
	if (grptr != grbuf)
		free(grptr);
	return status;
}
